#include <PID_v1.h>

//// hodnoty pro elektricke zapojeni
// piny na ktere je pripojen H-mustek zataceni
const int pinZataceniDoleva = 10;
const int pinZataceniDoprava = 11;
// stredni pin potenciometru zataceni
const byte pinPot = A0;

// piny pro H-mustky zataceni
// motory: cervena dopredu, zluta dozadu
// H-mustek leveho motoru
const int pinLevyDopredu = 6;
const int pinLevyDozadu = 9;
// H-mustek praveho motoru
const int pinPravyDopredu = 5;
const int pinPravyDozadu = 3;

/*
 * Zapojeni:
 * Zataceni:
 * bila+,zelena0: doleva
 * bila0,zelena+: doprava
 * 10 -- bila
 * 11 -- zelena
 * A0 -- hneda
 * 
 * Pohon:
 * cervena+, zluta0: dopredu
 * cervena0, zluta+: dozadu
 * L: 6 -- cervena, 9 -- bila
 * P: 5 -- cervena, 3 -- bila
 */


////// Ovladani zataceni //////
//// laditelne parametry zataceni ////

// kompenzace nulove polohy zataceni
double nuloveNatoceni = +0.010;

// krajni polohy zataceni
const float pravy_doraz = 0.560;
const float levy_doraz = -0.580;

// mrtve pasmo motoru
const float mrtve_pasmo = 0.040;

// tolerancni pasmo pro reg. odchylku
const float tolerancni_pasmo = 0.02;

// po jake minimalni dobe regulace prijme novou hodnotu
const int doba_regulace = 10; //ms

// parametry PID
float r0 = 3;
float rI = 1.0;
float rD = 0.0;

// filtrovani vstupu ... pocet vzorku pro prumerovani
const int pocet_vzorku = 25;
//// konec laditelnych parametru zataceni ////

////// Ovladani pohonu //////
//// laditelne parametry pohonu /////

// sekvence pro rozjezd ... omezeni zaberoveho proudu ... soft-start
const int pocetRychlosti = 10;
double rozjezdRychlost[pocetRychlosti] = {0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9};
int rozjezdCas[pocetRychlosti] = {10, 10, 30, 50, 50, 80, 80, 80, 100, 100};
//// konec laditelnych parametru pohonu ////

// inicializace promenny pro piny pohonu
int aktivniPin_P = 0;
int aktivniPin_L = 0;
int nulovyPin_P = 0;
int nulovyPin_L = 0;



/////// ostatni promenne ///////
// inicializace promennych
double rychlost = 0.0;
double rychlostPredchozi = 0.0;
double natoceni = 0.0;
double natoceniPredchozi = 0.0;

// doba na prijmuti novych hodnot
const int timeoutSerialu = 1000; // ms

// cas pro ukonceni neuspesne regulace
const int timeoutRegulace = 2000; // ms

// promenna pro ukladani casu behu programu
long int cas = 0;
long int f_cas = 0;

// nula pouzitelna s ukazatelem
double nula = 0.0;

//// inicializace PID regulatoru zataceni
// aktualni poloha, Input pro PID
double y = 0.0;
// zadana rychlost, Setpoint pro PID
double w = 0.0;
// akcni promenna, Output pro PID
double u = 0.0;
// objekt PID regulatoru
PID reg(&y, &u, &w, r0, rI, rD, DIRECT);

////// setup programu ///////
void setup() {
  Serial.begin(9600);
  // oznameni do seriove linky
  Serial.println("PID regulator pro zataceni + PWM rizeni dvou motoru");
  Serial.println("format: zataceni pohon");
  Serial.println("+0.420-0.230");
  // nastaveni vystupu
  pinMode(pinZataceniDoprava, OUTPUT);
  pinMode(pinZataceniDoleva, OUTPUT);

  // dioda na desce
  pinMode(13, OUTPUT);

  // nacteni hodnoty y
  y = map(analogRead(pinPot), 0, 1023, -1000, 1000) / 1000.0;
  for (int i = 0; i < pocet_vzorku - 1; i++) {
    y = y +  map(analogRead(pinPot), 0, 1023, -1000, 1000) / 1000.0;
  }
  y = y / pocet_vzorku;
  //Serial.println(y,3);
  // vychozi rychlost zadane hodnoty
  w = nuloveNatoceni;

  // nastaveni rozsahu akcni veliciny
  reg.SetOutputLimits(-1, 1);
  // prepnuti PID do automatickeho rezimu
  reg.SetMode(AUTOMATIC);


}

void loop() {
  // nacteni hodnoty z potenciometru:
  //    prevod vystupu ADC 0 az 1023 na -1.000 (vlevo) az 1.000 (vpravo)
  //    se zachovanim dostatecneho poctu desetinnych mist

  y = map(analogRead(pinPot), 0, 1023, -1000, 1000) / 1000.0;
  for (int i = 0; i < pocet_vzorku - 1; i++) {
    y = y +  map(analogRead(pinPot), 0, 1023, -1000, 1000) / 1000.0;
  }
  y = y / pocet_vzorku;

  // vypocteni u z aktualnich hodnot w a y
  reg.Compute();

  // zapsani hodnoty u na vystup
  zapis_u(u);

  // po dobe regulace a po dosazeni tolerancniho pasma prijme novou hodnotu
  if ((millis() - cas) > doba_regulace && (abs(w - y)) < tolerancni_pasmo) {
    prijmi_novy(natoceni, rychlost);
    w = natoceni;
    // pokud se rychlost zmenila, tak ji zapise
    if (rychlost != rychlostPredchozi){
      zapis_rychlost(rychlost, rychlostPredchozi);
    }
    rychlostPredchozi = rychlost;
    cas = millis();
  }

  // ukonceni regulace pri timeoutu
  if ((millis() - cas) > timeoutRegulace){
    zapis_rychlost(0, rychlostPredchozi);
    rychlostPredchozi = 0;
  }
}


void zapis_u(double &hodnota_u) {
  /* prevedeni hodnoty akcni hodnoty na odpovidajici vystup */

  // prevod 0 az 1 na rozsah 0 az 255 pro PWM
  int hodnota_pro_zapis = hodnota_u * 255;

  // vystup PWM signalu pro otaceni doprava nebo doleva
  if (abs(hodnota_pro_zapis) < mrtve_pasmo * 255) {
    // pri malem napeti motor neprekona mechanicky odpor
    analogWrite(pinZataceniDoleva, 0);
    analogWrite(pinZataceniDoprava, 0);
  }
  else {
    if (hodnota_pro_zapis > 0) {
      analogWrite(pinZataceniDoleva, 0);
      analogWrite(pinZataceniDoprava, hodnota_pro_zapis);
    }
    else {
      analogWrite(pinZataceniDoprava, 0);
      analogWrite(pinZataceniDoleva, -hodnota_pro_zapis);
    }
  }
}


void zapis_rychlost(double f_cilova, double f_predchozi) {
  if (f_cilova == 0){
    // zastavuje z predchozi rychlosti, tedy porty jsou spravne nastavene
    f_predchozi = abs(f_predchozi);
    zastaveni(f_predchozi);
    return;
  }

  // pokud maji ruzna znamenka, nejprve zastavi
  if (f_predchozi*f_cilova < 0){
    zastaveni(f_predchozi);
    f_predchozi=0;
    //debug?? zapis_rychlost(f_cilova,f_predchozi);
  }

  // nastaveni portu podle odpovidajiciho smeru
  if (f_cilova > 0){
     aktivniPin_P = pinPravyDopredu;
     aktivniPin_L = pinLevyDopredu;
     nulovyPin_P = pinPravyDozadu;
     nulovyPin_L = pinLevyDopredu;
  }
  if (f_cilova < 0){
    aktivniPin_P = pinPravyDozadu;
    aktivniPin_L = pinLevyDozadu;
    nulovyPin_P = pinPravyDopredu;
    nulovyPin_L = pinLevyDopredu;
  }

  rozjezd(f_cilova,f_predchozi);
  return;
}

void rozjezd(double f_rychlost, double f_predchoziRychlost) {
  if (f_rychlost < 0){
    /*
    aktivniPin_P = pinPravyDozadu;
    aktivniPin_L = pinLevyDozadu;
    nulovyPin_P = pinPravyDopredu;
    nulovyPin_L = pinLevyDopredu;*/
    f_rychlost = abs(f_rychlost);
  }

  analogWrite(nulovyPin_P,0);analogWrite(nulovyPin_L,0);

  // rozjezd na pozadovanou rychlost
  for (int i = 0; i < pocetRychlosti; i++) {
    // pokud je rychlost ze sekvence nizsi nez predchozi, preskoci se krok
    if (rozjezdRychlost[i]<f_predchoziRychlost){
      continue;
    }
    // pokud je rychlost ze sekvence vetsi, nez pozadovanan, cyklus se ukonci
    if (rozjezdRychlost[i] >= f_rychlost) {
      break;
    }
    // zapsani aktualniho kroku
    analogWrite(aktivniPin_P, rozjezdRychlost[i] * 255);
    analogWrite(aktivniPin_L, rozjezdRychlost[i] * 255);
    // prodleva prvku sekvence
    delay(rozjezdCas[i]);
  }

  // zapsani pozadovane rychlosti po rozjezdu
  analogWrite(aktivniPin_P, f_rychlost * 255);
  analogWrite(aktivniPin_L, f_rychlost * 255);
}

void zastaveni(double &f_predchozi) {
  // pokud zustava stat, pro jistotu znova zapise nulove hodnoty
  if (f_predchozi == 0){
  analogWrite(aktivniPin_P, 0);
  analogWrite(aktivniPin_L, 0);
  analogWrite(nulovyPin_P, 0);
  analogWrite(nulovyPin_L, 0);
  return;}


  if (f_predchozi < 0){
    /*
    aktivniPin_P = pinPravyDozadu;
    aktivniPin_L = pinLevyDozadu;
    nulovyPin_P = pinPravyDopredu;
    nulovyPin_L = pinLevyDopredu;*/
    f_predchozi = abs(f_predchozi);
  }

  // zpomaleni podle rozjezdove sekvence
  for (int i = pocetRychlosti - 1; i > 0; i--) {
    // pokud je rychlost vetsi, nez puvodni rychlost, preskoci se krok
    if (rozjezdRychlost[i] > f_predchozi) {
      continue;
    }
    // zapsani hodnoty rychlosti ze sekvence
    analogWrite(aktivniPin_P, rozjezdRychlost[i] * 255);
    analogWrite(aktivniPin_L, rozjezdRychlost[i] * 255);
    // prodleva
    delay(rozjezdCas[i]);
  }
  // finalni zastaveni
  analogWrite(aktivniPin_P, 0);
  analogWrite(aktivniPin_L, 0);
  analogWrite(nulovyPin_P, 0);
  analogWrite(nulovyPin_L, 0);
}



void prijmi_novy(double &natoceniNove, double &rychlostNova) {
  /*
    prijme novou hodnotu ze seriove linky.
    z prijatych znaku ve formatu char vytvori
    string a nasledne float
  */
  /* testovaci data:
     000000000000
     111111111111
     +0.250+0.250
     +0.500+0.500
     +1.000+1.000
     -0.250-0.250
     -1.000-1.000
     ------------
  */

  // motor zataceni bez napeti
  zapis_u(nula);

  // pocet znaku pro prijem, napr.: -0.124, +0.342
  const int pocet_znaku = 6;
  const int delka_zpravy = 12;

  bool prijatyNove = true;

  // string a float pro prijem ze seriove linky
  char prijem[delka_zpravy];
  char prijemNatoceni[pocet_znaku];
  char prijemRychlost[pocet_znaku];
  double prijemNatoceni_d = 0.0;
  double prijemRychlost_d = 0.0;

  // vycisteni vstupniho zasobniku seriove linky
  while (Serial.available() > 0) {
    Serial.read();
  }

  f_cas = millis();
  while (Serial.available() < delka_zpravy) {
    // ceka na prijmuti pocet_znaku znaku po vycisteni zasobniku
    // rozsviceni diody pri cekani
    digitalWrite(13, 1);

    // pokud nedostane nove hodnoty, tak nastavi na nulu
    if (millis() - f_cas > timeoutSerialu) {
      prijemNatoceni_d = 0.0;
      prijemRychlost_d = 0.0;
      prijatyNove = false;
      break;
    }
  }
  digitalWrite(13, 0);


  if (prijatyNove) {
    // prijmuti delka_zpravy znaku
    for (int i = 0; i < delka_zpravy; i++) {
      prijem[i] = Serial.read();
    }

    for (int i = 0; i < pocet_znaku; i++) {
      prijemNatoceni[i] = prijem[i];
    }

    for (int i = pocet_znaku; i < delka_zpravy; i++) {
      prijemRychlost[i - pocet_znaku] = prijem[i];
    }
    // prevod string na float
    prijemNatoceni_d = atof(prijemNatoceni);
    prijemRychlost_d = atof(prijemRychlost);

  }

  // omezeni maximalni hodnoty ~~ saturace
  if (prijemNatoceni_d > 1) {
    prijemNatoceni_d = 1.0;
  }
  if (prijemNatoceni_d < -1) {
    prijemNatoceni_d = -1.0;
  }
  if (prijemRychlost_d > 1) {
    prijemRychlost_d = 1.0;
  }
  if (prijemRychlost_d < -1) {
    prijemRychlost_d = -1.0;
  }

  // prepocet na dosazitelny rozsah
  if (prijemNatoceni_d > nuloveNatoceni) {
    prijemNatoceni_d = nuloveNatoceni + prijemNatoceni_d * (pravy_doraz - nuloveNatoceni);
  }
  else {
    prijemNatoceni_d = nuloveNatoceni - prijemNatoceni_d * (levy_doraz - nuloveNatoceni);
  }
  natoceniNove = prijemNatoceni_d;
  rychlostNova = prijemRychlost_d;

}
