#include <PID_v1.h>

// piny na ktere je pripojen H-mustek zataceni
const int levy_pin = 10;
const int pravy_pin = 11;
// stredni pin potenciometru zataceni
const byte pin = A0;


// kompenzace nulove polohy zataceni
double nulova_hodnota = +0.010;

// krajni polohy zataceni
const float pravy_doraz = 0.580;
const float levy_doraz = -0.600;


// mrtve pasmo motoru
const float mrtve_pasmo = 0.040;

// tolerancni pasmo pro reg. odchylku
const float tolerancni_pasmo = 0.01;

// po jake minimalni dobe regulace prijme novou hodnotu
const int doba_regulace = 100; //ms


// parametry PID
float r0 = 3;
float rI = 1.0;
float rD = 0.0;

// filtrovani vstupu
const int pocet_vzorku = 100;

// pocet znaku pro prijem, napr.: -0.124, +0.342
const int pocet_znaku = 6;

// promenna pro ukladani casu behu programu
long int cas=0;

// nula pouzitelna s ukazatelem
double nula = 0.0;

// string a float pro prijem ze seriove linky
char prijem[pocet_znaku];
double prijem_double=0.0;

// aktualni poloha, Input pro PID
double y = 0.0;
// zadana hodnota, Setpoint pro PID
double w = 0.0;
// akcni promenna, Output pro PID
double u = 0.0;


// inicializace PID regulatoru
PID reg(&y, &u, &w, r0, rI, rD, DIRECT);

void setup() {
  Serial.begin(9600);

  // nastaveni vystupu
  pinMode(pravy_pin, OUTPUT);
  pinMode(levy_pin, OUTPUT);

  // dioda na desce
  pinMode(13, OUTPUT);

  // nacteni hodnoty y 
  y = map(analogRead(pin), 0, 1023, -1000, 1000)/1000.0;
  for (int i = 0; i<pocet_vzorku-1; i++){
    y = y +  map(analogRead(pin), 0, 1023, -1000, 1000)/1000.0;
  }
  y = y / pocet_vzorku;
  //Serial.println(y,3);
  // vychozi hodnota zadane hodnoty
  w = nulova_hodnota;

  // nastaveni rozsahu akcni veliciny
  reg.SetOutputLimits(-1,1);
  // prepnuti PID do automatickeho rezimu
  reg.SetMode(AUTOMATIC);
}

void loop() {

  // nacteni hodnoty z potenciometru:
  //    prevod vystupu ADC 0 az 1023 na -1.000 (vlevo) az 1.000 (vpravo)
  //    se zachovanim dostatecneho poctu desetinnych mist
  y = map(analogRead(pin), 0, 1023, -1000, 1000)/1000.0;
  
  // vypocteni u z aktualnich hodnot w a y
  reg.Compute();

  // zapsani hodnoty u na vystup
  zapis_u(u);

  // po dobe regulace a po dosazeni tolerancniho pasma prijme novou hodnotu
  if ((millis()-cas)>doba_regulace && (abs(w-y))<tolerancni_pasmo){
    w = prijmi_novy();
    cas=millis();
  }

}


void zapis_u(double &hodnota_u){
  /* prevedeni hodnoty akcni hodnoty na odpovidajici vystup */

  // prevod 0 az 1 na rozsah 0 az 255 pro PWM
  int hodnota_pro_zapis = hodnota_u*255;

  // vystup PWM signalu pro otaceni doprava nebo doleva
  if (abs(hodnota_pro_zapis) < mrtve_pasmo*255){
    // pri malem napeti motor neprekona mechanicky odpor
    analogWrite(levy_pin, 0);
    analogWrite(pravy_pin, 0);
  }
  else {
    if (hodnota_pro_zapis > 0){
      analogWrite(levy_pin, 0);
      analogWrite(pravy_pin, hodnota_pro_zapis);
      }
    else {
      analogWrite(pravy_pin,0);
      analogWrite(levy_pin, -hodnota_pro_zapis);
    }
  }
}


float prijmi_novy(){
  /*
  prijme novou hodnotu ze seriove linky.
  z prijatych znaku ve formatu char vytvori
  string a nasledne float
  */

  // motor zataceni bez napeti
  zapis_u(nula);

  // vycisteni vstupniho zasobniku seriove linky
  while (Serial.available() > 0) {
  Serial.read();
  }

  while (Serial.available() < pocet_znaku){
    // ceka na prijmuti pocet_znaku znaku po vycisteni zasobniku
    // rozsviceni diody pri cekani
    digitalWrite(13,1);
  }
  digitalWrite(13,0);

  // prijmuti pocet_znaku znaku
  for (int i = 0; i < pocet_znaku; i++){
    prijem[i] = Serial.read();
  }

  // prevod string na float
  prijem_double=atof(prijem);
  
  // omezeni maximalni hodnoty ~~ saturace
  if (prijem_double>1){prijem_double=1.0;}
  if (prijem_double<-1){prijem_double=-1.0;}

  // prepocet na dosazitelny rozsah
  if (prijem_double > nulova_hodnota){
    prijem_double = nulova_hodnota + prijem_double * (pravy_doraz - nulova_hodnota);
  }
  else {
    prijem_double = nulova_hodnota - prijem_double * (levy_doraz - nulova_hodnota);
  }
  
  return(prijem_double);
}
